@@ -1,21 +1,59 @@
 use std::env;
 use std::fs::{self, File};
-use std::io;
 use std::io::prelude::*;
 use std::path::Path;
 
+use bstr::ByteSlice;
 use toml::Value;
 
 use crate::cmd::call;
 use crate::error::FatalError;
-use crate::Features;
+
+/// Expresses what features flags should be used
+pub enum Features {
+    /// None - don't use special features
+    None,
+    /// Only use selected features
+    Selective(Vec<String>),
+    /// Use all features via `all-features`
+    All,
+}
 
 fn cargo() -> String {
     env::var("CARGO").unwrap_or_else(|_| "cargo".to_owned())
 }
 
+pub fn package_content(manifest_path: &Path) -> Result<Vec<std::path::PathBuf>, FatalError> {
+    let mut cmd = std::process::Command::new(cargo());
+    cmd.arg("package");
+    cmd.arg("--manifest-path");
+    cmd.arg(manifest_path);
+    cmd.arg("--list");
+    // Not worth passing around allow_dirty to here since we are just getting a file list.
+    cmd.arg("--allow-dirty");
+    let output = cmd.output().map_err(FatalError::from)?;
+
+    let parent = manifest_path
+        .parent()
+        .unwrap_or_else(|| std::path::Path::new(""));
+
+    if output.status.success() {
+        let paths = ByteSlice::lines(output.stdout.as_slice())
+            .map(|l| parent.join(l.to_path_lossy()))
+            .collect();
+        Ok(paths)
+    } else {
+        let error = String::from_utf8_lossy(&output.stderr);
+        Err(FatalError::PackageListFailed(
+            manifest_path.to_owned(),
+            error.to_string(),
+        ))
+    }
+}
+
 pub fn publish(
     dry_run: bool,
+    no_verify: bool,
     manifest_path: &Path,
     features: &Features,
     registry: Option<&str>,
@@ -40,6 +78,15 @@ pub fn publish(
         command.push(token);
     }
 
+    if dry_run {
+        command.push("--dry-run");
+        command.push("--allow-dirty");
+    }
+
+    if no_verify {
+        command.push("--no-verify");
+    }
+
     let feature_arg;
     match features {
         Features::None => (),
@@ -53,7 +100,7 @@ pub fn publish(
         }
     };
 
-    call(command, dry_run)
+    call(command, false)
 }
 
 pub fn wait_for_publish(
@@ -101,7 +148,7 @@ pub fn set_package_version(manifest_path: &Path, version: &str) -> Result<(), Fa
         .join("Cargo.toml.work");
 
     {
-        let manifest = load_from_file(manifest_path)?;
+        let manifest = std::fs::read_to_string(manifest_path)?;
         let mut manifest: toml_edit::Document = manifest.parse().map_err(FatalError::from)?;
         manifest["package"]["version"] = toml_edit::value(version);
 
@@ -126,7 +173,7 @@ pub fn set_dependency_version(
         .join("Cargo.toml.work");
 
     {
-        let manifest = load_from_file(manifest_path)?;
+        let manifest = std::fs::read_to_string(manifest_path)?;
         let mut manifest: toml_edit::Document = manifest.parse().map_err(FatalError::from)?;
 
         let dep_table_names = &["dependencies", "dev-dependencies", "build-dependencies"];
@@ -179,12 +226,20 @@ fn set_version(
             return Ok(());
         }
     };
-    if dep_item.is_table_like() {
-        dep_item["version"] = toml_edit::value(version);
-    } else {
+
+    let mut is_table = false;
+    let mut has_version = false;
+    if let Some(table_like) = dep_item.as_table_like() {
+        is_table = true;
+        has_version = table_like.get("version").is_some();
+    }
+
+    if !is_table {
         return Err(FatalError::InvalidCargoFileFormat(
             "Intra-workspace dependencies should use both version and path".into(),
         ));
+    } else if has_version {
+        dep_item["version"] = toml_edit::value(version);
     }
 
     Ok(())
@@ -200,15 +255,58 @@ pub fn update_lock(manifest_path: &Path) -> Result<(), FatalError> {
 }
 
 pub fn parse_cargo_config(manifest_path: &Path) -> Result<Value, FatalError> {
-    let cargo_file_content = load_from_file(manifest_path).map_err(FatalError::from)?;
+    let cargo_file_content = std::fs::read_to_string(manifest_path).map_err(FatalError::from)?;
     cargo_file_content.parse().map_err(FatalError::from)
 }
 
-fn load_from_file(path: &Path) -> io::Result<String> {
-    let mut file = File::open(path)?;
-    let mut s = String::new();
-    file.read_to_string(&mut s)?;
-    Ok(s)
+pub fn sort_workspace(ws_meta: &cargo_metadata::Metadata) -> Vec<&cargo_metadata::PackageId> {
+    let members: std::collections::HashSet<_> = ws_meta.workspace_members.iter().collect();
+    let dep_tree: std::collections::HashMap<_, _> = ws_meta
+        .resolve
+        .as_ref()
+        .expect("cargo-metadata resolved deps")
+        .nodes
+        .iter()
+        .filter_map(|n| {
+            if members.contains(&n.id) {
+                Some((&n.id, &n.dependencies))
+            } else {
+                None
+            }
+        })
+        .collect();
+
+    let mut sorted = Vec::new();
+    let mut processed = std::collections::HashSet::new();
+    for pkg_id in ws_meta.workspace_members.iter() {
+        sort_workspace_inner(ws_meta, pkg_id, &dep_tree, &mut processed, &mut sorted);
+    }
+
+    sorted
+}
+
+fn sort_workspace_inner<'m>(
+    ws_meta: &'m cargo_metadata::Metadata,
+    pkg_id: &'m cargo_metadata::PackageId,
+    dep_tree: &std::collections::HashMap<
+        &'m cargo_metadata::PackageId,
+        &'m Vec<cargo_metadata::PackageId>,
+    >,
+    processed: &mut std::collections::HashSet<&'m cargo_metadata::PackageId>,
+    sorted: &mut Vec<&'m cargo_metadata::PackageId>,
+) {
+    if !processed.insert(pkg_id) {
+        return;
+    }
+
+    for dep_id in dep_tree[pkg_id]
+        .iter()
+        .filter(|dep_id| dep_tree.contains_key(dep_id))
+    {
+        sort_workspace_inner(ws_meta, dep_id, dep_tree, processed, sorted);
+    }
+
+    sorted.push(pkg_id);
 }
 
 #[cfg(test)]
@@ -283,7 +381,7 @@ mod test {
             set_dependency_version(manifest_path.path(), "foo", "2.0").unwrap();
 
             manifest_path.assert(
-                predicate::str::similar(
+                predicate::str::diff(
                     r#"
     [package]
     name = "t"
@@ -329,7 +427,7 @@ mod test {
             set_dependency_version(manifest_path.path(), "foo", "2.0").unwrap();
 
             manifest_path.assert(
-                predicate::str::similar(
+                predicate::str::diff(
                     r#"
     [package]
     name = "t"
@@ -375,7 +473,7 @@ mod test {
             set_dependency_version(manifest_path.path(), "foo", "2.0").unwrap();
 
             manifest_path.assert(
-                predicate::str::similar(
+                predicate::str::diff(
                     r#"
     [package]
     name = "t"
@@ -421,7 +519,7 @@ mod test {
             set_dependency_version(manifest_path.path(), "foo", "2.0").unwrap();
 
             manifest_path.assert(
-                predicate::str::similar(
+                predicate::str::diff(
                     r#"
     [package]
     name = "t"
@@ -471,7 +569,7 @@ mod test {
             set_dependency_version(manifest_path.path(), "foo", "2.0").unwrap();
 
             manifest_path.assert(
-                predicate::str::similar(
+                predicate::str::diff(
                     r#"
     [package]
     name = "t"
@@ -524,6 +622,48 @@ mod test {
             temp.close().unwrap();
         }
 
+        #[test]
+        fn no_version() {
+            let temp = assert_fs::TempDir::new().unwrap();
+            temp.copy_from("tests/fixtures/simple", &["**"]).unwrap();
+            let manifest_path = temp.child("Cargo.toml");
+            manifest_path
+                .write_str(
+                    r#"
+    [package]
+    name = "t"
+    version = "0.1.0"
+    authors = []
+    edition = "2018"
+
+    [dependencies]
+    foo = { path = "../" }
+    "#,
+                )
+                .unwrap();
+
+            set_dependency_version(manifest_path.path(), "foo", "2.0").unwrap();
+
+            manifest_path.assert(
+                predicate::str::diff(
+                    r#"
+    [package]
+    name = "t"
+    version = "0.1.0"
+    authors = []
+    edition = "2018"
+
+    [dependencies]
+    foo = { path = "../" }
+    "#,
+                )
+                .from_utf8()
+                .from_file_path(),
+            );
+
+            temp.close().unwrap();
+        }
+
         #[test]
         fn out_of_line_table() {
             let temp = assert_fs::TempDir::new().unwrap();
@@ -550,7 +690,7 @@ mod test {
             set_dependency_version(manifest_path.path(), "foo", "2.0").unwrap();
 
             manifest_path.assert(
-                predicate::str::similar(
+                predicate::str::diff(
                     r#"
     [package]
     name = "t"
